from typing import List
from pathlib import Path
import srsly
from .types import Example
from .validation import json_to_examples


def read_jsonl(path: Path) -> List[Example]:
    """Read a jsonl annotations file 
    
    ### Parameters
    --------------
    **path**: (Path), required.
        Path to data
    
    ### Returns
    -----------
    (List[Example]): 
        List of Examples
    """
    
    data = list(srsly.read_jsonl(path))
    return json_to_examples(data)


def read_json(path: Path) -> List[Example]:
    """Read a json annotations file 
    
    ### Parameters
    --------------
    **path**: (Path), required.
        Path to data
    
    ### Returns
    -----------
    (List[Example]): 
        List of Examples
    """
    data = srsly.read_json(path)
    return json_to_examples(data)



from reconner.stats import ner_stats


def test_ner_stats(example_data):
    stats = ner_stats(example_data['train'])
    printed_stats = ner_stats(example_data['train'], serialize=True)
    str_stats = ner_stats(example_data['train'], serialize=True, no_print=True)
    assert printed_stats is None
    assert isinstance(str_stats, str)

    assert stats['n_examples'] == 102
    assert stats['n_examples_no_entities'] == 29
    assert stats['ents_per_type'] == {
        "SKILL": 191,
        "PRODUCT": 34,
        "JOB_ROLE": 5
    }


import pytest
from reconner.recognizer import *


def test_base_recognizer(test_texts):
    recognizer = EntityRecognizer()

    with pytest.raises(NotImplementedError):
        recognizer.labels
    with pytest.raises(NotImplementedError):
        recognizer.predict(test_texts)
            

def test_spacy_recognizer(nlp, test_texts):
    ruler = nlp.create_pipe('entity_ruler')
    ruler.add_patterns([
        {'label': 'SKILL', 'pattern': 'Machine learning'},
        {'label': 'SKILL', 'pattern': 'researched'},
        {'label': 'SKILL', 'pattern': 'AI'},
        {'label': 'JOB_ROLE', 'pattern': 'Software Engineer'},
    ])

    nlp.add_pipe(ruler)

    recognizer = SpacyEntityRecognizer(nlp)
    assert recognizer.labels == ["JOB_ROLE", "SKILL"]

    examples = recognizer.predict(test_texts)

    assert examples[0].text == test_texts[0]
    assert len(examples[0].spans) == 3
    assert len(examples[1].spans) == 2



from pathlib import Path
from typing import Any, Callable, Dict, List
from reconner.loaders import read_jsonl, read_json
from reconner.types import Example


class Dataset:
    """Container for a full dataset with train/dev/test splits.
    Used to apply core functions to all datasets at once.

    ### Parameters
    --------------
    **train**: (List[Example]), required.
        List of Examples for **train** set
    **dev**: (List[Example]), required.
        List of Examples for **dev** set
    **test**: (List[Example], optional), Defaults to None.
        List of Examples for **test** set
    """
    
    def __init__(self,
                 train: List[Example],
                 dev: List[Example],
                 test: List[Example] = None):
        self.datasets = {
            'train': train,
            'dev': dev,
            'all': train + dev
        }
        if test:
            self.datasets.update({
                'test': test,
                'all': train + dev + test
            })
    
    @classmethod
    def from_disk(cls, path: Path, loader_func: Callable = read_jsonl):
        """Load Dataset from disk given a directory with files 
        named explicitly train.jsonl, dev.jsonl, and test.jsonl
        
        ### Parameters
        --------------
        **path**: (Path), required.
            directory to load from
        **loader_func**: (Callable, optional), Defaults to read_jsonl.
            Loader function (TODO: Make this a bit more generic)
        """
        return Dataset(
            loader_func(path / 'train.jsonl'),
            loader_func(path / 'dev.jsonl'),
            test=loader_func(path / 'test.jsonl')
        )

    def apply(self,
              func: Callable[[List[Example]], Any],
              *args: Any,
              **kwargs: Any) -> Dict[str, List[Example]]:
        """Apply an existing function to all datasets
        
        ### Parameters
        --------------
        **func**: (Callable[[List[Example]], Any]), required.
            Function from an existing reconner module that can operate on a List of Examples
        
        ### Returns
        -----------
        (Dict[str, List[Example]]): 
            Dictionary mapping dataset names to List[Example], same as the internal datasets property
        """
        res = {}
        for k, dataset in self.datasets.items():
            res[k] = func(dataset, *args, **kwargs)
        return res



from typing import List
from pydantic import StrictStr
from spacy.language import Language
from .types import Example, TextSpanLabel


class EntityRecognizer:

    @property
    def labels(self) -> List[str]:
        """Return List of String Labels
        
        ### Raises
        -----------
        NotImplementedError: 
            Not Implemented, override
        
        ### Returns
        -----------
        (List[str]): 
            List of labels the model can predict
        """
        raise NotImplementedError

    def predict(self, texts: List[StrictStr]) -> List[Example]:
        """Run model inference on a batch of raw texts.
        
        ### Parameters
        --------------
        **texts**: (List[StrictStr]), required.
            Raw text examples
        
        ### Raises
        -----------
        NotImplementedError: 
            Not implemented, override
        
        ### Returns
        -----------
        (List[Example]): 
            List of Examples constructed from Model predictions
        """
        raise NotImplementedError


class SpacyEntityRecognizer(EntityRecognizer):

    def __init__(self, nlp: Language):
        super().__init__()
        self.nlp = nlp
    
    @property
    def labels(self) -> List[str]:
        """Return List of spaCy ner labels
        
        ### Returns
        -----------
        (List[str]): 
            List of labels from spaCy ner pipe
        """
        all_labels = set()

        for pipe in ['ner', 'entity_ruler']:
            if self.nlp.has_pipe(pipe):
                all_labels = all_labels | set(self.nlp.get_pipe(pipe).labels)

        return sorted(list(all_labels))


    def predict(self, texts: List[StrictStr]) -> List[Example]:
        """Run spaCy nlp.pipe on a batch of raw texts.
        
        ### Parameters
        --------------
        **texts**: (List[StrictStr]), required.
            Raw text examples
        
        ### Returns
        -----------
        (List[Example]): 
            List of Examples constructed from spaCy Model predictions
        """
        examples: List[Example] = []

        for doc in self.nlp.pipe(texts):
            examples.append(Example(
                text=doc.text,
                spans=[TextSpanLabel(text=e.text,
                                     start=e.start_char,
                                     end=e.end_char,
                                     label=e.label_)
                                     for e in doc.ents]
            ))
        return examples



from markdown import Markdown
from markdown.extensions import Extension
from markdown.blockprocessors import BlockProcessor
from markdown.util import etree
import importlib
import inspect
import re
import typing


# Fuzzy regex for determining source lines in __init__ that look like
# attribute assignments.  Eg. `self.counter = 0`
SET_ATTRIBUTE = re.compile("^([ \t]*)self[.]([A-Za-z0-9_]+) *=")


def import_from_string(import_str: str) -> typing.Any:
    module_str, _, attr_str = import_str.rpartition(".")

    try:
        module = importlib.import_module(module_str)
    except ImportError as exc:
        module_name = module_str.split(".", 1)[0]
        if exc.name != module_name:
            raise exc from None
        raise ValueError(f"Could not import module {module_str!r}.")

    try:
        return getattr(module, attr_str)
    except AttributeError as exc:
        raise ValueError(f"Attribute {attr_str!r} not found in module {module_str!r}.")


def get_params(signature: inspect.Signature) -> typing.List[str]:
    """
    Given a function signature, return a list of parameter strings
    to use in documentation.
    Eg. test(a, b=None, **kwargs) -> ['a', 'b=None', '**kwargs']
    """
    params = []
    render_pos_only_separator = True
    render_kw_only_separator = True

    for parameter in signature.parameters.values():
        value = parameter.name
        if parameter.default is not parameter.empty:
            value = f"{value}={parameter.default!r}"

        if parameter.kind is parameter.VAR_POSITIONAL:
            render_kw_only_separator = False
            value = f"*{value}"
        elif parameter.kind is parameter.VAR_KEYWORD:
            value = f"**{value}"
        elif parameter.kind is parameter.POSITIONAL_ONLY:
            if render_pos_only_separator:
                render_pos_only_separator = False
                params.append("/")
        elif parameter.kind is parameter.KEYWORD_ONLY:
            if render_kw_only_separator:
                render_kw_only_separator = False
                params.append("*")
        params.append(value)

    return params


def last_iter(seq: typing.Sequence) -> typing.Iterator:
    """
    Given a sequence, return a two-tuple (item, is_last) iterable.
    See: https://stackoverflow.com/a/1633483/596689
    """
    it = iter(seq)
    item = next(it)
    is_last = False

    for next_item in it:
        yield item, is_last
        item = next_item

    is_last = True
    yield item, is_last


def trim_docstring(docstring: typing.Optional[str]) -> str:
    """
    Trim leading indent from a docstring.
    See: https://www.python.org/dev/peps/pep-0257/#handling-docstring-indentation
    """
    if not docstring:
        return ""

    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = 1000
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))

    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < 1000:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())

    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)

    # Return a single string:
    return "\n".join(trimmed[:trimmed.index('### Parameters')]) + "\n\n".join(trimmed[trimmed.index('### Parameters'):])


class AutoDocProcessor(BlockProcessor):

    CLASSNAME = "autodoc"
    RE = re.compile(r"(?:^|\n)::: ?([:a-zA-Z0-9_.]*) *(?:\n|$)")
    RE_SPACES = re.compile("  +")

    def __init__(self, parser, md=None):
        super().__init__(parser=parser)
        self.md = md

    def test(self, parent: etree.Element, block: etree.Element) -> bool:
        sibling = self.lastChild(parent)
        return bool(
            self.RE.search(block)
            or (
                block.startswith(" " * self.tab_length)
                and sibling is not None
                and sibling.get("class", "").find(self.CLASSNAME) != -1
            )
        )

    def run(self, parent: etree.Element, blocks: etree.Element) -> None:
        sibling = self.lastChild(parent)
        block = blocks.pop(0)
        m = self.RE.search(block)

        if m:
            block = block[m.end() :]  # removes the first line

        block, theRest = self.detab(block)

        if m:
            import_string = m.group(1)
            item = import_from_string(import_string)

            autodoc_div = etree.SubElement(parent, "div")
            autodoc_div.set("class", self.CLASSNAME)

            self.render_signature(autodoc_div, item, import_string)
            for line in block.splitlines():
                if line.startswith(":docstring:"):
                    docstring = trim_docstring(item.__doc__)
                    self.render_docstring(autodoc_div, item, docstring)
                elif line.startswith(":members:"):
                    members = line.split()[1:] or None
                    self.render_members(autodoc_div, item, members=members)

        if theRest:
            # This block contained unindented line(s) after the first indented
            # line. Insert these lines as the first block of the master blocks
            # list for future processing.
            blocks.insert(0, theRest)

    def render_signature(
        self, elem: etree.Element, item: typing.Any, import_string: str
    ) -> None:
        module_string, _, name_string = import_string.rpartition(".")

        # Eg: `some_module.attribute_name`
        signature_elem = etree.SubElement(elem, "div")
        signature_elem.set("class", "autodoc-signature")

        if inspect.isclass(item):
            qualifier_elem = etree.SubElement(signature_elem, "em")
            qualifier_elem.text = "class "
        elif inspect.iscoroutinefunction(item):
            qualifier_elem = etree.SubElement(signature_elem, "em")
            qualifier_elem.text = "async "

        name_elem = etree.SubElement(signature_elem, "code")
        if module_string:
            name_elem.text = module_string + "."
        main_name_elem = etree.SubElement(name_elem, "strong")
        main_name_elem.text = name_string

        # If this is a property, then we're done.
        if not callable(item):
            return

        # Eg: `(a, b='default', **kwargs)``
        signature = inspect.signature(item)

        bracket_elem = etree.SubElement(signature_elem, "span")
        bracket_elem.text = "("
        bracket_elem.set("class", "autodoc-punctuation")

        if signature.parameters:
            for param, is_last in last_iter(get_params(signature)):
                param_elem = etree.SubElement(signature_elem, "em")
                param_elem.text = param
                param_elem.set("class", "autodoc-param")

                if not is_last:
                    comma_elem = etree.SubElement(signature_elem, "span")
                    comma_elem.text = ", "
                    comma_elem.set("class", "autodoc-punctuation")

        bracket_elem = etree.SubElement(signature_elem, "span")
        bracket_elem.text = ")"
        bracket_elem.set("class", "autodoc-punctuation")

    def render_docstring(
        self, elem: etree.Element, item: typing.Any, docstring: str
    ) -> None:
        docstring_elem = etree.SubElement(elem, "div")
        docstring_elem.set("class", "autodoc-docstring")

        md = Markdown(extensions=self.md.registeredExtensions)
        docstring_elem.text = md.convert(docstring)

    def render_members(
        self, elem: etree.Element, item: typing.Any, members: typing.List[str] = None
    ) -> None:
        members_elem = etree.SubElement(elem, "div")
        members_elem.set("class", "autodoc-members")

        if members is None:
            members = sorted([attr for attr in dir(item) if not attr.startswith("_")])

        info_items = []
        for attribute_name in members:
            attribute = getattr(item, attribute_name)
            docs = trim_docstring(getattr(attribute, "__doc__", ""))
            info = (attribute_name, docs)
            info_items.append(info)

        for attribute_name, docs in info_items:
            attribute = getattr(item, attribute_name)
            self.render_signature(members_elem, attribute, attribute_name)
            self.render_docstring(members_elem, attribute, docs)


class MKAutoDocExtension(Extension):
    def extendMarkdown(self, md: Markdown) -> None:
        md.registerExtension(self)
        processor = AutoDocProcessor(md.parser, md=md)
        md.parser.blockprocessors.register(processor, "mkautodoc", 110)


def makeExtension():
    return MKAutoDocExtension()



from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Optional
import srsly

from .constants import NONE
from .types import Example


def ner_stats(data: List[Example],
              serialize: bool = False,
              no_print: bool = False) -> Optional[List[Example]]:
    """Compute statistics for NER data
    
    ### Parameters
    --------------
    **data**: (List[Example]), required.
        Data as a List of Examples
    **serialize**: (bool, optional), Defaults to False.
        Serialize to a JSON string for printing
    **no_print**: (bool, optional), Defaults to False.
        Don't print, return serialized string. Requires serialize to be True
    
    ### Returns
    -----------
    (List[Example]): 
        List of examples or string if serialize and no_print are both True
    """    
    labels: DefaultDict[str, Any] = defaultdict(int)
    examples: DefaultDict[str, Any] = defaultdict(list)
    n_examples_no_entities = 0
    for e in data:
        if not e.spans:
            n_examples_no_entities += 1
            examples[NONE].append(e)
        else:
            for s in e.spans:
                labels[s.label] += 1
                examples[s.label].append(e)

    res = {
        'n_examples': len(data),
        'n_examples_no_entities': n_examples_no_entities,
        'ents_per_type': labels
    }
    if serialize:
        s = srsly.json_dumps(res, indent=4)
        if no_print:
            return s
        else:
            print(s)
    else:
        res['examples_with_type'] = examples
        return res


from pathlib import Path
from reconner.dataset import Dataset
from reconner.stats import ner_stats


def test_dataset_initialize(example_data):
    ds1 = Dataset(example_data['train'], example_data['dev'])
    assert len(ds1.datasets['train']) == 102
    assert len(ds1.datasets['dev']) == 110
    assert len(ds1.datasets['all']) == 212

    ds2 = Dataset(example_data['train'], example_data['dev'], test=example_data['test'])
    assert len(ds2.datasets['train']) == 102
    assert len(ds2.datasets['dev']) == 110
    assert len(ds2.datasets['test']) == 96
    assert len(ds2.datasets['all']) == 308


def test_dataset_from_disk(example_data):
    ds = Dataset.from_disk(Path(__file__).parent.parent / 'examples/data')
    assert ds.datasets['train'] == example_data['train']
    assert ds.datasets['dev'] == example_data['dev']
    assert ds.datasets['test'] == example_data['test']
    assert len(ds.datasets['all']) == 308


def test_dataset_apply(example_data):
    dataset = Dataset(example_data['train'], example_data['dev'], test=example_data['test'])
    stats = dataset.apply(ner_stats)
    assert sorted(list(stats.keys())) == ['all', 'dev', 'test', 'train']

    assert stats['all']['n_examples'] == 308



"""ReconNER, Debug annotated Named Entity Recognition (NER) data for inconsitencies and get insights on improving the quality of your data."""

__version__ = "0.0.1"

from .dataset import *
from .insights import *
from .stats import *
from .validation import *


from typing import Any, Dict, List, Tuple
from .types import Example


def json_to_examples(data: List[Dict[str, Any]]) -> List[Example]:
    """Convert List of Dicts to List of typed Examples
    
    ### Parameters
    --------------
    **data**: (List[Dict[str, Any]]), required.
        Input List of Dicts to convert
    
    ### Returns
    -----------
    (List[Example]): 
        List of typed Examples
    """    
    data = fix_annotations_format(data)
    data = filter_overlaps(data)
    return [Example(**example) for example in data]


def fix_annotations_format(data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Fix annotations format for a consistent dataset
    
    ### Parameters
    --------------
    **data**: (List[Dict[str, Any]]), required.
        List of Examples
    
    ### Returns
    -----------
    (List[Dict[str, Any]]): 
        List of Examples with corrected formatting
    """    
    for e in data:
        if 'meta' not in e:
            e['meta'] = {}
        if isinstance(e['meta'], list) or isinstance(e['meta'], str):
            e['meta'] = {
                'source': e['meta']
            }

        for s in e['spans']:
            if 'text' not in s:
                s['text'] = e['text'][s['start']:s['end']]
            s['label'] = s['label'].upper()
    return data


def filter_overlaps(data: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Filter overlapping entity spans by picking the longest one.
    
    ### Parameters
    --------------
    **data**: (List[Dict[str, Any]]), required.
        List of Examples
    
    ### Returns
    -----------
    (List[Dict[str, Any]]): 
        List of Examples with fixed overlaps
    """
    out_data = []
    for e in data:
        annotations = sorted(e['spans'], key=lambda x: x['start'])
        filtered_annotations = remove_overlapping_entities(annotations)
        
        new_e = e.copy()
        new_e['spans'] = filtered_annotations
        out_data.append(new_e)

    return out_data


def select_subset_of_overlapping_chain(
    chain: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    """
    Select the subset of entities in an overlapping chain to return by greedily choosing the
    longest entity in the chain until there are no entities remaining
    """
    sorted_chain = sorted(chain, key=lambda x: x['end'] - x['start'], reverse=True)
    selections_from_chain = []
    chain_index = 0
    # dump the current chain by greedily keeping the longest entity that doesn't overlap
    while chain_index < len(sorted_chain):
        entity = sorted_chain[chain_index]
        match_found = False
        for already_selected_entity in selections_from_chain:
            max_start = max(entity['start'], already_selected_entity['start'])
            min_end = min(entity['end'], already_selected_entity['end'])
            if len(range(max_start, min_end)) > 0:
                match_found = True
                break

        if not match_found:
            selections_from_chain.append(entity)

        chain_index += 1

    return selections_from_chain


def remove_overlapping_entities(
    sorted_spacy_format_entities: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    """
    Removes overlapping entities from the entity set, by greedilytaking the longest
    entity from each overlapping chain. The input list of entities should be sorted
    and follow the spacy format.
    """
    spacy_format_entities_without_overlap = []
    current_overlapping_chain: List[Dict[str, Any]] = []
    current_overlapping_chain_start = 0
    current_overlapping_chain_end = 0
    for i, current_entity in enumerate(sorted_spacy_format_entities):
        current_entity = sorted_spacy_format_entities[i]
        current_entity_start = current_entity['start']
        current_entity_end = current_entity['end']

        if len(current_overlapping_chain) == 0:
            current_overlapping_chain.append(current_entity)
            current_overlapping_chain_start = current_entity_start
            current_overlapping_chain_end = current_entity_end
        else:
            min_end = min(current_entity_end, current_overlapping_chain_end)
            max_start = max(current_entity_start, current_overlapping_chain_start)
            if min_end - max_start > 0:
                current_overlapping_chain.append(current_entity)
                current_overlapping_chain_start = min(
                    current_entity_start, current_overlapping_chain_start
                )
                current_overlapping_chain_end = max(
                    current_entity_end, current_overlapping_chain_end
                )
            else:
                selections_from_chain = select_subset_of_overlapping_chain(
                    current_overlapping_chain
                )

                current_overlapping_chain = []
                spacy_format_entities_without_overlap.extend(selections_from_chain)
                current_overlapping_chain.append(current_entity)
                current_overlapping_chain_start = current_entity_start
                current_overlapping_chain_end = current_entity_end

    spacy_format_entities_without_overlap.extend(
        select_subset_of_overlapping_chain(current_overlapping_chain)
    )

    return sorted(spacy_format_entities_without_overlap, key=lambda x: x['start'])



from typing import Any, Dict, List
from pydantic import validator, BaseModel, Field, Schema
from pydantic import StrictStr, StrictInt, StrictFloat, StrictBool


class TextSpanLabel(BaseModel):
    text: StrictStr
    start: StrictInt
    end: StrictInt
    label: StrictStr


class Example(BaseModel):
    text: StrictStr
    spans: List[TextSpanLabel]
    meta: Dict[str, Any] = {}


class PredictionError(BaseModel):
    text: StrictStr
    true_label: StrictStr
    false_label: StrictStr
    count: StrictInt
